from asyncio import create_task, gather, get_event_loop, sleep
from contextlib import suppress
from functools import lru_cache
from typing import Any, Dict, Optional, Tuple, Union

import evmspec
import y._db.common
import y._db.config as config
from a_sync import PruningThreadPoolExecutor, a_sync
from brownie import chain
from eth_typing import ChecksumAddress
from evmspec.data import _decode_hook
from logging import getLogger
from msgspec import ValidationError, json
from multicall.utils import get_event_loop
from pony.orm import (
    BindingError,
    OperationalError,
    TransactionIntegrityError,
    commit,
    db_session,
    flush,
    select,
)
from y import ENVIRONMENT_VARIABLES as ENVS
from y._db.entities import db
from y.constants import CHAINID
from y.exceptions import reraise_excs_with_extra_context

from eth_portfolio._db import entities
from eth_portfolio._db.decorators import break_locks, requery_objs_on_diff_tx_err
from eth_portfolio._db.entities import (
    AddressExtended,
    BlockExtended,
    ContractExtended,
    TokenExtended,
)
from eth_portfolio._decimal import Decimal
from eth_portfolio.structs import InternalTransfer, TokenTransfer, Transaction, TransactionRLP
from eth_portfolio.typing import _P, _T, Fn

logger = getLogger(__name__)


def __bind():
    try:
        db.bind(**config.connection_settings)
    except BindingError as e:
        if not str(e).startswith("Database object was already bound to"):
            raise e


__bind()

try:
    db.generate_mapping(create_tables=True)
except OperationalError as e:
    if not str(e).startswith("no such column:"):
        raise
    raise OperationalError(
        "Since eth-portfolio extends the ypricemagic database with additional column definitions, you will need to delete your ypricemagic database at ~/.ypricemagic and rerun this script"
    ) from e

from y._db.decorators import retry_locked
from y._db.entities import Address, Block, Chain, Contract, Token, insert

# The db must be bound before we do this since we're adding some new columns to the tables defined in ypricemagic
from y._db.utils import ensure_chain, get_chain
from y._db.utils.price import _set_price
from y._db.utils.traces import insert_trace
from y import ERC20
from y.constants import EEE_ADDRESS
from y.exceptions import NonStandardERC20
from y.contracts import is_contract


_big_pool_size = 4 if ENVS.DB_PROVIDER == "sqlite" else 10
_small_pool_size = 2 if ENVS.DB_PROVIDER == "sqlite" else 4

_block_executor = PruningThreadPoolExecutor(_small_pool_size, "eth-portfolio block")
_token_executor = PruningThreadPoolExecutor(_small_pool_size, "eth-portfolio token")
_address_executor = PruningThreadPoolExecutor(_small_pool_size, "eth-portfolio address")
_transaction_read_executor = PruningThreadPoolExecutor(
    _big_pool_size, "eth-portfolio-transaction-read"
)
_transaction_write_executor = PruningThreadPoolExecutor(
    _small_pool_size, "eth-portfolio-transaction-write"
)
_token_transfer_read_executor = PruningThreadPoolExecutor(
    _big_pool_size, "eth-portfolio-token-transfer-read"
)
_token_transfer_write_executor = PruningThreadPoolExecutor(
    _small_pool_size, "eth-portfolio-token-transfer-write"
)
_internal_transfer_read_executor = PruningThreadPoolExecutor(
    _big_pool_size, "eth-portfolio-internal-transfer read"
)
_internal_transfer_write_executor = PruningThreadPoolExecutor(
    _small_pool_size, "eth-portfolio-internal-transfer write"
)


def robust_db_session(fn: Fn[_P, _T]) -> Fn[_P, _T]:
    return retry_locked(break_locks(db_session(fn)))


db_session_cached = lambda func: retry_locked(
    lru_cache(maxsize=None)(db_session(retry_locked(func)))
)


@a_sync(default="async", executor=_block_executor)
@robust_db_session
def get_block(block: int) -> BlockExtended:
    if b := BlockExtended.get(chain=CHAINID, number=block):
        return b
    elif b := Block.get(chain=CHAINID, number=block):
        if isinstance(b, BlockExtended):
            # in case of race cndtn
            return b
        raise ValueError(b, b.number, b.CHAINID)
        hash = b.hash
        ts = b.timestamp
        prices = [(price.token.address, price.price) for price in b.prices]
        logs = [json.decode(log.raw) for log in b.logs]
        traces = [json.decode(trace.raw) for trace in b.traces]
        for p in b.prices:
            p.delete()
        for l in b.logs:
            l.delete()
        for t in b.traces:
            t.delete()
        flush()
        b.delete()
        commit()
        b = insert(
            type=BlockExtended,
            chain=get_chain(sync=True),
            number=block,
            hash=hash,
            timestamp=ts,
        )
        try:
            # for log in logs:
            #    insert_log(log)
            for trace in traces:
                insert_trace(trace)
        except Exception as e:
            e.args = (
                "This is really bad. Might need to nuke your db if you value your logs/traces",
                *e.args,
            )
            raise
        for token, price in prices:
            _set_price(token, price, sync=True)
    asdasd = get_chain(sync=True)
    if not isinstance(asdasd, Chain):
        raise TypeError(asdasd)
    commit()
    if b := insert(type=BlockExtended, chain=asdasd, number=block):
        return b
    return BlockExtended.get(chain=CHAINID, number=block)


@a_sync(default="async", executor=_block_executor)
@db_session_cached
def ensure_block(block: int) -> None:
    get_block(block, sync=True)


# TODO refactor this out, async is annoying sometimes
# process = ProcessPoolExecutor(
#  1,
#  # NOTE: come on apple, what are you dooooin?
#  mp_context=get_context('fork'),
# )


def is_token(address) -> bool:
    if address == EEE_ADDRESS:
        return False
    # with suppress(NonStandardERC20):
    #    erc = ERC20(address)
    #    if all(erc.symbol, erc.name, erc.total_supply(), erc.scale):
    #    #if all(erc._symbol(), erc._name(), erc.total_supply(), erc._scale()):
    #        return True
    # return False
    return get_event_loop().run_until_complete(_is_token(address))


async def _is_token(address) -> bool:
    # just breaking a weird lock, dont mind me
    if retval := await get_event_loop().run_in_executor(process, __is_token, address):
        logger.debug("%s is token")
    else:
        logger.debug("%s is not token")
    return retval


def __is_token(address) -> bool:
    with suppress(NonStandardERC20):
        erc = ERC20(address, asynchronous=True)
        if all(
            get_event_loop().run_until_complete(
                gather(erc._symbol(), erc._name(), erc.total_supply_readable())
            )
        ):
            return True
    return False


@a_sync(default="async", executor=_address_executor)
@robust_db_session
def get_address(address: ChecksumAddress) -> AddressExtended:
    entity_type = TokenExtended
    entity = entities.Address.get(chain=CHAINID, address=address)
    """ TODO: fix this later
    entity = entities.Address.get(chain=chain, address=address)
    if isinstance(entity, (Token, TokenExtended)):
        entity_type = TokenExtended
    elif isinstance(entity, (Contract, ContractExtended)):
        entity_type = ContractExtended
    elif isinstance(entity, (Address, AddressExtended)):
        entity_type = AddressExtended
    elif entity is None:
        # TODO: this logic should live in ypm, prob
        entity_type = AddressExtended if not is_contract(address) else TokenExtended if is_token(address) else ContractExtended
    else:
        raise NotImplementedError(entity, entity_type)
        
    if isinstance(entity, entity_type):
        return entity
    
    elif entity:
        logger.debug("deleting %s", entity)
        entity.delete()
        commit()
    """
    if entity := entities.Address.get(chain=CHAINID, address=address):
        return entity

    ensure_chain()
    return insert(type=entity_type, chain=CHAINID, address=address) or entity_type.get(
        chain=CHAINID, address=address
    )


@a_sync(default="async", executor=_address_executor)
@db_session_cached
def ensure_address(address: ChecksumAddress) -> None:
    get_address(address, sync=True)


@a_sync(default="async", executor=_address_executor)
@db_session_cached
def ensure_addresses(*addresses: ChecksumAddress) -> None:
    for address in addresses:
        ensure_address(address, sync=True)


@a_sync(default="async", executor=_token_executor)
@robust_db_session
def get_token(address: ChecksumAddress) -> TokenExtended:
    if t := TokenExtended.get(chain=CHAINID, address=address):
        return t
    kwargs = {}
    if t := Address.get(chain=CHAINID, address=address):
        if isinstance(t, TokenExtended):
            # double check due to possible race cntdn
            return t
        """
        with suppress(TypeError):
            if t.notes:
                kwargs['notes'] = t.notes
        if isinstance(t, Contract):
            with suppress(TypeError):
                if t.deployer:
                    kwargs['deployer'] = t.deployer
            with suppress(TypeError):
                if t.deploy_block:
                    kwargs['deploy_block'] = t.deploy_block
        if isinstance(t, Token):
            with suppress(TypeError):
                if t.symbol:
                    kwargs['symbol'] = t.symbol
            with suppress(TypeError):
                if t.name:
                    kwargs['name'] = t.name
            with suppress(TypeError):
                if t.bucket:
                    kwargs['bucket'] = t.bucket
        """

        try:
            flush()
            t.delete()
            commit()
        except KeyError as e:
            raise KeyError(f"cant delete {t}") from e

    ensure_chain()
    commit()
    return insert(
        type=TokenExtended, chain=CHAINID, address=address, **kwargs
    ) or TokenExtended.get(chain=CHAINID, address=address)


@a_sync(default="async", executor=_token_executor)
@db_session_cached
def ensure_token(token_address: ChecksumAddress) -> None:
    get_token(token_address, sync=True)


async def get_transaction(sender: ChecksumAddress, nonce: int) -> Optional[Transaction]:
    startup_txs = await transactions_known_at_startup(CHAINID, sender)
    data = startup_txs.pop(nonce, None) or await __get_transaction_bytes_from_db(sender, nonce)
    if data:
        await _yield_to_loop()
        return decode_transaction(data)


_decoded = 0


async def _yield_to_loop():
    """dont let the event loop get congested, let your rpc begin work asap"""
    global _decoded
    _decoded += 1
    if _decoded % 1000 == 0:
        await sleep(0)


@a_sync(default="async", executor=_transaction_read_executor)
@robust_db_session
def __get_transaction_bytes_from_db(sender: ChecksumAddress, nonce: int) -> Optional[bytes]:
    entity: entities.Transaction
    if entity := entities.Transaction.get(from_address=(CHAINID, sender), nonce=nonce):
        return entity.raw


def decode_transaction(data: bytes) -> Union[Transaction, TransactionRLP]:
    try:
        try:
            return json.decode(data, type=Transaction, dec_hook=_decode_hook)
        except ValidationError as e:
            if str(e) == "Object missing required field `type` - at `$[2]`":
                return json.decode(data, type=TransactionRLP, dec_hook=_decode_hook)
            raise
    except Exception as e:
        e.args = *e.args, json.decode(data)
        raise


@a_sync(default="async", executor=_transaction_write_executor)
@robust_db_session
def delete_transaction(transaction: Transaction) -> None:
    if entity := entities.Transaction.get(**transaction.__db_primary_key__):
        entity.delete()


async def insert_transaction(transaction: Transaction) -> None:
    # Make sure these are in the db so below we can call them and use the results all in one transaction
    # NOTE: this create task -> await coro -> await task pattern is faster than a 2-task gather
    block_task = create_task(ensure_block(transaction.block_number))
    if to_address := transaction.to_address:
        await ensure_addresses(to_address, transaction.from_address)
    else:
        await ensure_address(transaction.from_address)
    await block_task
    await _insert_transaction(transaction)


@a_sync(default="async", executor=_transaction_write_executor)
@requery_objs_on_diff_tx_err
@robust_db_session
def _insert_transaction(transaction: Transaction) -> None:
    with reraise_excs_with_extra_context(transaction):
        entities.Transaction(
            **transaction.__db_primary_key__,
            block=(CHAINID, transaction.block_number),
            transaction_index=transaction.transaction_index,
            hash=transaction.hash.hex(),
            to_address=(CHAINID, transaction.to_address) if transaction.to_address else None,
            value=transaction.value,
            price=transaction.price,
            value_usd=transaction.value_usd,
            type=getattr(transaction, "type", None),
            gas=transaction.gas,
            gas_price=transaction.gas_price,
            max_fee_per_gas=getattr(transaction, "max_fee_per_gas", None),
            max_priority_fee_per_gas=getattr(transaction, "max_priority_fee_per_gas", None),
            raw=json.encode(transaction, enc_hook=enc_hook),
        )


@a_sync(default="async", executor=_internal_transfer_read_executor)
@robust_db_session
def get_internal_transfer(trace: evmspec.FilterTrace) -> Optional[InternalTransfer]:
    block = trace.blockNumber
    entity: entities.InternalTransfer
    if entity := entities.InternalTransfer.get(
        block=(CHAINID, block),
        transaction_index=trace.transactionPosition,
        hash=trace.transactionHash,
        type=trace.type.name,
        call_type=trace.callType,
        from_address=(CHAINID, trace.sender),
        to_address=(CHAINID, trace.to),
        value=trace.value.scaled,
        trace_address=(CHAINID, trace.traceAddress),
        gas=trace.gas,
        gas_used=trace.gasUsed if "gasUsed" in trace else None,
        input=trace.input,
        output=trace.output,
        subtraces=trace.subtraces,
        address=(CHAINID, trace.address),
    ):
        return json.decode(entity.raw, type=InternalTransfer, dec_hook=_decode_hook)


@a_sync(default="async", executor=_internal_transfer_write_executor)
@robust_db_session
def delete_internal_transfer(transfer: InternalTransfer) -> None:
    if entity := entities.InternalTransfer.get(
        block=(CHAINID, transfer.block_number),
        transaction_index=transfer.transaction_index,
        hash=transfer.hash,
        type=transfer.type,
        call_type=transfer.call_type,
        from_address=(CHAINID, transfer.from_address),
        to_address=(CHAINID, transfer.to_address),
        value=transfer.value,
        trace_address=(CHAINID, transfer.trace_address),
        gas=transfer.gas,
        gas_used=transfer.gas_used,
        input=transfer.input,
        output=transfer.output,
        subtraces=transfer.subtraces,
        address=(CHAINID, transfer.address),
    ):
        entity.delete()


async def insert_internal_transfer(transfer: InternalTransfer) -> None:
    # NOTE: this create task -> await coro -> await task pattern is faster than a 2-task gather
    block_task = create_task(ensure_block(transfer.block_number))
    if to_address := getattr(transfer, "to_address", None):
        await ensure_addresses(to_address, transfer.from_address)
    else:
        await ensure_address(transfer.from_address)
    await block_task
    await _insert_internal_transfer(transfer)


@a_sync(default="async", executor=_internal_transfer_write_executor)
@robust_db_session
def _insert_internal_transfer(transfer: InternalTransfer) -> None:
    entities.InternalTransfer(
        block=(CHAINID, transfer.block_number),
        transaction_index=transfer.transaction_index,
        hash=transfer.hash,
        type=transfer.type,
        call_type=transfer.call_type,
        from_address=(CHAINID, transfer.from_address),
        to_address=(CHAINID, transfer.to_address),
        value=transfer.value,
        price=transfer.price,
        value_usd=transfer.value_usd,
        trace_address=str(transfer.trace_address),
        gas=transfer.gas,
        gas_used=transfer.gas_used,
        raw=json.encode(transfer, enc_hook=enc_hook),
    )


async def get_token_transfer(transfer: evmspec.Log) -> Optional[TokenTransfer]:
    pk = {
        "block": (CHAINID, transfer.blockNumber),
        "transaction_index": transfer.transactionIndex,
        "log_index": transfer.logIndex,
    }
    startup_xfers = await token_transfers_known_at_startup(CHAINID)
    data = startup_xfers.pop(tuple(pk.values()), None) or await __get_token_transfer_bytes_from_db(
        pk
    )
    if data:
        await _yield_to_loop()
        with reraise_excs_with_extra_context(data):
            return json.decode(data, type=TokenTransfer, dec_hook=_decode_hook)


@a_sync(default="async", executor=_token_transfer_read_executor)
@robust_db_session
def __get_token_transfer_bytes_from_db(pk: dict) -> Optional[bytes]:
    entity: entities.TokenTransfer
    if entity := entities.TokenTransfer.get(**pk):
        return entity.raw


_TPK = Tuple[Tuple[int, ChecksumAddress], int]


@a_sync(default="async", executor=_transaction_read_executor, ram_cache_maxsize=None)
@robust_db_session
def transactions_known_at_startup(chainid: int, from_address: ChecksumAddress) -> Dict[_TPK, bytes]:
    return dict(
        select(
            (t.nonce, t.raw)
            for t in entities.Transaction  # type: ignore [attr-defined]
            if t.from_address.chain.id == chainid and t.from_address.address == from_address
        )
    )


_TokenTransferPK = Tuple[Tuple[int, int], int, int]


@a_sync(default="async", executor=_transaction_read_executor, ram_cache_maxsize=None)
@robust_db_session
def token_transfers_known_at_startup(chainid: int) -> Dict[_TokenTransferPK, bytes]:
    block: int
    tx_index: int
    log_index: int
    raw: bytes

    transfers = {}
    for block, tx_index, log_index, raw in select(
        (t.block.number, t.transaction_index, t.log_index, t.raw)
        for t in entities.TokenTransfer  # type: ignore [attr-defined]
        if t.block.chain.id == chainid
    ):
        pk = ((chainid, block), tx_index, log_index)
        transfers[pk] = raw
    return transfers


@a_sync(default="async", executor=_token_transfer_write_executor)
@robust_db_session
def delete_token_transfer(token_transfer: TokenTransfer) -> None:
    if entity := entities.TokenTransfer.get(
        block=(CHAINID, token_transfer.block_number),
        transaction_index=token_transfer.transaction_index,
        log_index=token_transfer.log_index,
    ):
        entity.delete()


async def insert_token_transfer(token_transfer: TokenTransfer) -> None:
    # two tasks and a coroutine like this should be faster than gather
    block_task = create_task(ensure_block(token_transfer.block_number))
    while True:
        try:
            token_task = create_task(ensure_token(token_transfer.token_address))
        except KeyError:
            # This KeyError comes from a bug in cachetools.ttl_cache
            # TODO: move this handler into evmspec
            pass
        else:
            break

    while True:
        try:
            address_coro = ensure_addresses(token_transfer.to_address, token_transfer.from_address)
        except KeyError:
            # This KeyError comes from a bug in cachetools.ttl_cache
            # TODO: move this handler into evmspec
            pass
        else:
            break

    await address_coro
    await block_task
    await token_task
    await _insert_token_transfer(token_transfer)


@a_sync(default="async", executor=_token_transfer_write_executor)
@requery_objs_on_diff_tx_err
@robust_db_session
def _insert_token_transfer(token_transfer: TokenTransfer) -> None:
    try:
        entities.TokenTransfer(
            block=(CHAINID, token_transfer.block_number),
            transaction_index=token_transfer.transaction_index,
            log_index=token_transfer.log_index,
            hash=token_transfer.hash.hex(),
            token=(CHAINID, token_transfer.token_address),
            from_address=(CHAINID, token_transfer.from_address),
            to_address=(CHAINID, token_transfer.to_address),
            value=token_transfer.value,
            price=token_transfer.price,
            value_usd=token_transfer.value_usd,
            raw=json.encode(token_transfer, enc_hook=enc_hook),
        )
        commit()
    except TransactionIntegrityError:
        pass  # most likely non-issue, debug later if needed


def enc_hook(obj: Any) -> Any:
    try:
        return y._db.common.enc_hook(obj)
    except TypeError:
        if type(obj) is Decimal:
            return obj.jsonify()
        raise TypeError(type(obj), obj) from None



"""
This module defines the :class:`~eth_portfolio.Portfolio` and :class:`~eth_portfolio.PortfolioWallets` classes, which are used to manage and interact with a collection of :class:`~eth_portfolio.address.PortfolioAddress` objects.
It also includes the :class:`~eth_portfolio.PortfolioLedger` class for handling transactions, internal transfers, and token transfers associated with the portfolio.
The :class:`~eth_portfolio.Portfolio` class leverages the `a_sync` library to support both synchronous and asynchronous operations. This allows it to efficiently gather data, perform portfolio-related tasks, and handle network calls without blocking, thus improving the overall responsiveness and performance of portfolio operations.

This file is part of a larger system that includes modules for handling portfolio addresses, ledger entries, and other related tasks.
"""

import logging
from functools import wraps
from typing import Any, AsyncIterator, Dict, Iterable, Iterator, List, Optional, Tuple, Union

import a_sync
from a_sync import igather
from a_sync.a_sync import ASyncFunction
from brownie import web3
from checksum_dict import ChecksumAddressDict
from pandas import DataFrame, concat  # type: ignore
from web3 import Web3
from y.datatypes import Address, Block

from eth_portfolio import _argspec
from eth_portfolio._decorators import set_end_block_if_none
from eth_portfolio._ledgers.address import PandableLedgerEntryList
from eth_portfolio._ledgers.portfolio import (
    PortfolioInternalTransfersLedger,
    PortfolioLedgerBase,
    PortfolioTokenTransfersLedger,
    PortfolioTransactionsLedger,
)
from eth_portfolio._utils import _LedgeredBase
from eth_portfolio.address import PortfolioAddress
from eth_portfolio.constants import ADDRESSES
from eth_portfolio.structs import LedgerEntry
from eth_portfolio.typing import Addresses, PortfolioBalances

logger = logging.getLogger(__name__)


class PortfolioWallets(Iterable[PortfolioAddress], Dict[Address, PortfolioAddress]):  # type: ignore [misc]
    """
    A container that holds all :class:`~eth_portfolio.address.PortfolioAddress` objects for a specific :class:`~eth_portfolio.Portfolio`.

    Works like a ``Dict[Address, PortfolioAddress]`` except when you iterate you get the values instead of the keys.
    You should not initialize these. They are created automatically during :class:`~eth_portfolio.Portfolio` initialization.

    Attributes:
        _wallets: A checksummed dictionary of :class:`~eth_portfolio.address.PortfolioAddress` objects.
    """

    _wallets: ChecksumAddressDict[PortfolioAddress]

    def __init__(
        self,
        addresses: Iterable[Address],
        start_block: Block,
        load_prices: bool,
        num_workers_transactions: int,
    ) -> None:
        """
        Initialize a PortfolioWallets instance.

        Args:
            portfolio: The :class:`~eth_portfolio.Portfolio` instance to which this wallet belongs.
            addresses: An iterable of addresses to be included in the portfolio.
        """
        self._wallets: ChecksumAddressDict[PortfolioAddress] = ChecksumAddressDict()
        """
        A checksummed dictionary of :class:`~eth_portfolio.address.PortfolioAddress` objects.

        Type: 
            ChecksumAddressDict[PortfolioAddress]
        """

        for address in addresses:
            self._wallets[address] = PortfolioAddress(
                address,
                start_block,
                load_prices,
                num_workers_transactions=num_workers_transactions,
                asynchronous=portfolio.asynchronous,
            )

    def __repr__(self) -> str:
        """
        Return a string representation of the PortfolioWallets instance.

        Returns:
            String representation of the instance.
        """
        return f"<{type(self).__name__} wallets={list(self._wallets.values())}>"

    def __contains__(self, address: Union[Address, PortfolioAddress]) -> bool:
        """
        Check if an address is in the portfolio wallets.

        Args:
            address: The address to check.

        Returns:
            True if the address is in the wallets, False otherwise.
        """
        return address in self._wallets

    def __getitem__(self, address: Address) -> PortfolioAddress:
        """
        Get the :class:`~eth_portfolio.address.PortfolioAddress` object for a given address.

        Args:
            address: The address to look up.

        Returns:
            PortfolioAddress: The :class:`~eth_portfolio.address.PortfolioAddress` object.
        """
        return self._wallets[address]

    def __iter__(self) -> Iterator[PortfolioAddress]:
        """
        Iterate over the :class:`~eth_portfolio.address.PortfolioAddress` objects in the wallets.

        Returns:
            Iterator[PortfolioAddress]: An iterator over :class:`~eth_portfolio.address.PortfolioAddress` objects.
        """
        yield from self._wallets.values()

    def __len__(self) -> int:
        """
        Get the number of :class:`~eth_portfolio.address.PortfolioAddress` objects in the wallets.

        Returns:
            The number of :class:`~eth_portfolio.address.PortfolioAddress` objects.
        """
        return len(self._wallets)

    def __bool__(self) -> bool:
        """
        Check if the wallets contain any addresses.

        Returns:
            True if there are addresses in the wallets, False otherwise.
        """
        return bool(self._wallets)

    def keys(self) -> Iterable[Address]:
        """
        Get the keys (addresses) of the wallets.

        Returns:
            Iterable[Address]: An iterable of addresses.
        """
        return self._wallets.keys()

    def values(self) -> Iterable[PortfolioAddress]:
        """
        Get the values (:class:`~eth_portfolio.address.PortfolioAddress` objects) of the wallets.

        Returns:
            Iterable[PortfolioAddress]: An iterable of :class:`~eth_portfolio.address.PortfolioAddress` objects.
        """
        return self._wallets.values()

    def items(self) -> Iterable[Tuple[Address, PortfolioAddress]]:
        """
        Get the items (address, :class:`~eth_portfolio.address.PortfolioAddress` pairs) of the wallets.

        Returns:
            Iterable[Tuple[Address, PortfolioAddress]]: An iterable of (address, :class:`~eth_portfolio.address.PortfolioAddress`) tuples.
        """
        return self._wallets.items()


_DEFAULT_LABEL = "your portfolio"


class Portfolio(a_sync.ASyncGenericBase):
    """
    Used to export information about a group of :class:`~eth_portfolio.address.PortfolioAddress` objects.

    - Has all attributes of a :class:`~eth_portfolio.address.PortfolioAddress`.
    - All calls to `function(*args, **kwargs)` will return `{address: PortfolioAddress(Address).function(*args, **kwargs)}`
    """

    label: str = _DEFAULT_LABEL
    """
    A label for the portfolio. Defaults to "your portfolio"
    """

    load_prices: bool = True
    """
    Whether to load prices. Defaults to True.
    """

    asynchronous: bool = False
    """
    Whether to use asynchronous operations. Defaults to False.
    """

    _start_block = 0
    """
    The starting block number. Defaults to 0.
    """

    def __init__(
        self,
        addresses: Addresses,
        start_block: int = 0,
        label: str = _DEFAULT_LABEL,
        load_prices: bool = True,
        num_workers_transactions: int = 1000,
        asynchronous: bool = False,
    ) -> None:
        """
        Initialize a Portfolio instance.

        Args:
            addresses: The addresses to include in the portfolio.
        Examples:
            >>> portfolio = Portfolio(addresses=["0xAddress1", "0xAddress2"])
            >>> print(portfolio)
        """
        if not isinstance(start_block, int):
            raise TypeError(f"`start_block` must be an integer, not {type(start_block)}")
        if start_block < 0:
            raise ValueError("`start_block` must be >= 0")
        super().__init__()

        if start_block:
            self._start_block = start_block

        assert isinstance(label, str), f"`label` must be a string, you passed {type(label)}"
        if label != _DEFAULT_LABEL:
            self.label = label

        if load_prices is False:
            self.load_prices = False
        elif load_prices is not True:
            raise TypeError(f"`load_prices` must be a boolean, you passed {type(load_prices)}")

        if asynchronous is True:
            self.asynchronous = True
        elif asynchronous is not False:
            raise TypeError(f"`asynchronous` must be a boolean, you passed {type(asynchronous)}")

        if isinstance(addresses, str):
            addresses = [addresses]
        elif not isinstance(addresses, Iterable):
            raise TypeError(f"`addresses` must be an iterable, not {type(addresses)}")

        self.addresses = PortfolioWallets(
            addresses, start_block, load_prices, num_workers_transactions
        )
        """
        A container for the :class:`~eth_portfolio.Portfolio`'s :class:`~eth_portfolio.address.PortfolioAddress` objects.
        
        Type: 
            :class:`~eth_portfolio.PortfolioWallets`
        
        Works like a ``Dict[Address, PortfolioAddress]`` except you get the values when you iterate instead of the keys.
        """

        self.ledger = PortfolioLedger(self)
        """
        A container for all of your fun taxable events.
        
        Type: 
            :class:`~eth_portfolio.PortfolioLedger`
        """

        self.w3: Web3 = web3
        """
        The `Web3` object which will be used to call your rpc for all read operations.
        
        Type: 
            :class:`~web3.Web3`
        """

    def __repr__(self) -> str:
        # NOTE: should we show addresses? for now we won't
        # string = f"{type(self).__name__}(addresses={list(map(str, self.addresses))}"
        string = f"{type(self).__name__}(label={self.label}"
        if self._start_block:
            string += f", start_block={self._start_block}"
        return string + f", load_prices={self.load_prices}, asynchronous={self.asynchronous})"

    def __repr__(self) -> str:
        # NOTE: should we show addresses? for now we won't
        # string = f"{type(self).__name__}(addresses={list(map(str, self.addresses))}"
        string = f"{type(self).__name__}(label={self.label}"
        if self._start_block:
            string += f", start_block={self._start_block}"
        return string + f", load_prices={self.load_prices}, asynchronous={self.asynchronous})"

    def __getitem__(self, key: Address) -> PortfolioAddress:
        """
        Get a :class:`~eth_portfolio.address.PortfolioAddress` by its key.

        Args:
            key: The address key.

        Returns:
            PortfolioAddress: The :class:`~eth_portfolio.address.PortfolioAddress` object.

        Example:
            >>> portfolio = Portfolio(addresses=["0xAddress1", "0xAddress2"])
            >>> address = portfolio["0xAddress1"]
            >>> print(address)
        """
        return self.addresses[key]

    def __iter__(self) -> Iterator[PortfolioAddress]:
        """
        Iterate over the :class:`~eth_portfolio.address.PortfolioAddress` objects.

        Returns:
            Iterator[PortfolioAddress]: An iterator over :class:`~eth_portfolio.address.PortfolioAddress` objects.

        Example:
            >>> portfolio = Portfolio(addresses=["0xAddress1", "0xAddress2"])
            >>> for address in portfolio:
            ...     print(address)
        """
        yield from self.addresses

    @property
    def transactions(self) -> PortfolioTransactionsLedger:
        """
        A container for all transactions to or from any of your :class:`~eth_portfolio.address.PortfolioAddress`.

        Returns:
            PortfolioTransactionsLedger: The :class:`~eth_portfolio._ledgers.portfolio.PortfolioTransactionsLedger` object.

        Example:
            >>> portfolio = Portfolio(addresses=["0xAddress1", "0xAddress2"])
            >>> transactions = portfolio.transactions
            >>> print(transactions)
        """
        return self.ledger.transactions

    @property
    def internal_transfers(self) -> PortfolioInternalTransfersLedger:
        """
        A container for all internal transfers to or from any of your :class:`~eth_portfolio.address.PortfolioAddress`.

        Returns:
            PortfolioInternalTransfersLedger: The :class:`~eth_portfolio._ledgers.portfolio.PortfolioInternalTransfersLedger` object.

        Example:
            >>> portfolio = Portfolio(addresses=["0xAddress1", "0xAddress2"])
            >>> internal_transfers = portfolio.internal_transfers
            >>> print(internal_transfers)
        """
        return self.ledger.internal_transfers

    @property
    def token_transfers(self) -> PortfolioTokenTransfersLedger:
        """
        A container for all token transfers to or from any of your :class:`~eth_portfolio.address.PortfolioAddress`.

        Returns:
            PortfolioTokenTransfersLedger: The :class:`~eth_portfolio._ledgers.portfolio.PortfolioTokenTransfersLedger` object.

        Example:
            >>> portfolio = Portfolio(addresses=["0xAddress1", "0xAddress2"])
            >>> token_transfers = portfolio.token_transfers
            >>> print(token_transfers)
        """
        return self.ledger.token_transfers

    async def describe(self, block: int) -> PortfolioBalances:
        """
        Returns a full snapshot of your portfolio at a given block.

        Args:
            block: The block number.

        Returns:
            PortfolioBalances: A snapshot of the portfolio balances.

        Example:
            >>> portfolio = Portfolio(addresses=["0xAddress1", "0xAddress2"])
            >>> balances = await portfolio.describe(block=1200000)
            >>> print(balances)
        """
        assert block
        return PortfolioBalances(
            await a_sync.gather(
                {address.address: address.describe(block, sync=False) for address in self}
            ),
            block=block,
        )

    async def sent(
        self,
        start_block: Optional[Block] = None,
        end_block: Optional[Block] = None,
        *,
        mem_cache: bool = True,
    ) -> AsyncIterator[LedgerEntry]:
        async for obj in self.ledger.sent(start_block, end_block, mem_cache=mem_cache):
            yield obj

    async def received(
        self,
        start_block: Optional[Block] = None,
        end_block: Optional[Block] = None,
        *,
        mem_cache: bool = True,
    ) -> AsyncIterator[LedgerEntry]:
        async for obj in self.ledger.received(start_block, end_block, mem_cache=mem_cache):
            yield obj


async_functions = {
    name: obj for name, obj in PortfolioAddress.__dict__.items() if isinstance(obj, ASyncFunction)
}
for func_name, func in async_functions.items():
    if not callable(getattr(PortfolioAddress, func_name)):
        raise RuntimeError(
            f"A PortfolioAddress object should not have a non-callable attribute suffixed with '_async'"
        )

    @a_sync.a_sync(default=func.default)
    @wraps(func)
    async def imported_func(self: Portfolio, *args: Any, **kwargs: Any) -> _argspec.get_return_type(getattr(PortfolioAddress, func_name)):  # type: ignore
        """
        Import an asynchronous function from :class:`~eth_portfolio.address.PortfolioAddress` to :class:`~eth_portfolio.Portfolio`.

        Args:
            self: The :class:`~eth_portfolio.Portfolio` instance.
            args: Positional arguments for the function.
            kwargs: Keyword arguments for the function.

        Returns:
            Any: The return type of the function.

        Example:
            >>> portfolio = Portfolio(addresses=["0xAddress1", "0xAddress2"])
            >>> result = await portfolio.some_async_function(*args, **kwargs)
            >>> print(result)
        """
        return await a_sync.gather(
            {address: func(address, *args, **kwargs, sync=False) for address in self}
        )

    setattr(Portfolio, func_name, imported_func)
    logger.debug("Ported %s from PortfolioAddress to Portfolio", func_name)


def _get_missing_cols_from_KeyError(e: KeyError) -> List[str]:
    """
    Extract missing column names from a KeyError.

    Args:
        e: The KeyError exception.

    Returns:
        A list of missing column names.
    """
    split = str(e).split("'")
    return [split[i * 2 + 1] for i in range(len(split) // 2)]


class PortfolioLedger(_LedgeredBase[PortfolioLedgerBase]):
    """
    A container for all transactions, internal transfers, and token transfers to or from any of the wallets in your :class:`~eth_portfolio.Portfolio`.
    """

    def __init__(self, portfolio: "Portfolio") -> None:
        """
        Initialize a PortfolioLedger instance.

        Args:
            portfolio: The :class:`~eth_portfolio.Portfolio` instance to which this ledger belongs.
        """
        super().__init__(portfolio._start_block)
        self.portfolio = portfolio
        """
        The :class:`~eth_portfolio.Portfolio` containing the wallets this ledger will pertain to.
        """
        self.transactions = PortfolioTransactionsLedger(portfolio)
        """
        A container for all transactions to or from any of your :class:`~eth_portfolio.address.PortfolioAddress`.
        """
        self.internal_transfers = PortfolioInternalTransfersLedger(portfolio)
        """
        A container for all internal transfers to or from any of your :class:`~eth_portfolio.address.PortfolioAddress`.
        """
        self.token_transfers = PortfolioTokenTransfersLedger(portfolio)
        """
        A container for all token transfers to or from any of your :class:`~eth_portfolio.address.PortfolioAddress`.
        """
        self.asynchronous = portfolio.asynchronous
        """
        True if default mode is async, False if sync.
        """

    async def all_entries(
        self, start_block: Block, end_block: Block
    ) -> Dict[PortfolioAddress, Dict[str, PandableLedgerEntryList]]:
        """
        Returns a mapping containing all transactions, internal transfers, and token transfers to or from each wallet in your portfolio.

        Args:
            start_block: The starting block number.
            end_block: The ending block number.

        Returns:
            Dict[PortfolioAddress, Dict[str, PandableLedgerEntryList]]: A dictionary mapping :class:`~eth_portfolio.address.PortfolioAddress` to their ledger entries.

        Example:
            >>> portfolio = Portfolio(addresses=["0xAddress1", "0xAddress2"])
            >>> entries = await portfolio.ledger.all_entries(start_block=1000000, end_block=1100000)
            >>> print(entries)
        """
        return await a_sync.gather(
            {address: address.all(start_block, end_block, sync=False) for address in self.portfolio}
        )

    @set_end_block_if_none
    async def df(self, start_block: Block, end_block: Block, full: bool = False) -> DataFrame:
        """
        Returns a DataFrame containing all transactions, internal transfers, and token transfers to or from any wallet in your portfolio.

        Args:
            start_block: The starting block number.
            end_block: The ending block number.
            full (optional): Whether to include all columns or a subset. Defaults to False.

        Returns:
            DataFrame: A DataFrame with the ledger entries.

        Example:
            >>> portfolio = Portfolio(addresses=["0xAddress1", "0xAddress2"])
            >>> df = await portfolio.ledger.df(start_block=1000000, end_block=1100000)
            >>> print(df)
        """
        df = concat(
            await igather(ledger.df(start_block, end_block, sync=False) for ledger in self._ledgers)
        )

        # Reorder columns
        while True:
            try:
                if full:
                    df = df[
                        [
                            "chainId",
                            "blockNumber",
                            "blockHash",
                            "transactionIndex",
                            "hash",
                            "log_index",
                            "nonce",
                            "from",
                            "to",
                            "token",
                            "token_address",
                            "value",
                            "price",
                            "value_usd",
                            "gas",
                            "gasPrice",
                            "gasUsed",
                            "maxFeePerGas",
                            "maxPriorityFeePerGas",
                            "type",
                            "callType",
                            "traceAddress",
                            "subtraces",
                            "output",
                            "error",
                            "result",
                            "address",
                            "code",
                            "init",
                            "r",
                            "s",
                            "v",
                            "input",
                        ]
                    ]
                else:
                    df = df[
                        [
                            "chainId",
                            "blockNumber",
                            "hash",
                            "from",
                            "to",
                            "token",
                            "token_address",
                            "value",
                            "price",
                            "value_usd",
                            "gas",
                            "gasPrice",
                            "gasUsed",
                            "maxFeePerGas",
                            "maxPriorityFeePerGas",
                            "type",
                            "callType",
                            "traceAddress",
                            "subtraces",
                            "output",
                            "error",
                            "result",
                            "address",
                        ]
                    ]
                    df = df[df["value"] != 0]
                break
            except KeyError as e:
                for column in _get_missing_cols_from_KeyError(e):
                    df[column] = None

        sort_cols = (
            ["blockNumber", "transactionIndex", "log_index"] if full else ["blockNumber", "hash"]
        )

        try:
            return df.sort_values(sort_cols).reset_index(drop=True)
        except KeyError:
            logger.error(df)
            logger.error(df.columns)
            raise

    async def sent(
        self,
        start_block: Optional[Block] = None,
        end_block: Optional[Block] = None,
        *,
        mem_cache: bool = True,
    ) -> AsyncIterator[LedgerEntry]:
        portfolio_addresses = set(self.portfolio.addresses.keys())
        async for obj in self._get_and_yield(start_block or 0, end_block, mem_cache):
            if (
                obj.value
                and obj.from_address in portfolio_addresses
                and obj.to_address not in portfolio_addresses
            ):
                yield obj

    async def received(
        self,
        start_block: Optional[Block] = None,
        end_block: Optional[Block] = None,
        *,
        mem_cache: bool = True,
    ) -> AsyncIterator[LedgerEntry]:
        portfolio_addresses = set(self.portfolio.addresses.keys())
        async for obj in self._get_and_yield(start_block or 0, end_block, mem_cache):
            if (
                obj.value
                and obj.to_address in portfolio_addresses
                and obj.from_address not in portfolio_addresses
            ):
                yield obj


# Use this var for a convenient way to set up your portfolio using env vars.
portfolio = Portfolio(ADDRESSES)



import os
import sys

from brownie import network

sys.path.insert(0, os.path.abspath("."))

try:
    brownie_network = os.environ["PYTEST_NETWORK"]
except KeyError:
    raise ValueError(
        "Please set the PYTEST_NETWORK environment variable to the name of the brownie network you want to use for testing."
    )

if not network.is_connected():
    network.connect(brownie_network)

"""
This module configures the testing environment for the project by ensuring
that the Brownie network is connected using the specified network. It also
modifies the system path to include the current directory, allowing for
importing modules from the project root.

Environment Variables:
    PYTEST_NETWORK: The name of the Brownie network to use for testing. This
    environment variable must be set before running the tests.

Raises:
    ValueError: If the PYTEST_NETWORK environment variable is not set.

Examples:
    To set the PYTEST_NETWORK environment variable and run tests:

    .. code-block:: bash

        export PYTEST_NETWORK=mainnet-fork
        pytest

    This will connect to the specified Brownie network and run the tests.

See Also:
    - :mod:`brownie.network`: For more information on managing network connections with Brownie.
    - :mod:`os`: For more information on interacting with the operating system.
    - :mod:`sys`: For more information on the system-specific parameters and functions.
"""



from typing import List, Optional

import a_sync
from y.datatypes import Address, Block

from eth_portfolio._submodules import import_submodules, get_protocols
from eth_portfolio.protocols import lending
from eth_portfolio.protocols._base import StakingPoolABC
from eth_portfolio.typing import RemoteTokenBalances


import_submodules()

protocols: List[StakingPoolABC] = get_protocols()  # type: ignore [assignment]


@a_sync.future
async def balances(address: Address, block: Optional[Block] = None) -> RemoteTokenBalances:
    """
    Fetch token balances for a given address across various protocols.

    This function retrieves the token balances for a specified Ethereum address
    at a given block across all available protocols. It is decorated with
    :func:`a_sync.future`, allowing it to be used in both synchronous and
    asynchronous contexts.

    If no protocols are available, the function returns an empty
    :class:`~eth_portfolio.typing.RemoteTokenBalances` object.

    Args:
        address: The Ethereum address for which to fetch balances.
        block: The block number at which to fetch balances.
            If not provided, the latest block is used.

    Examples:
        Fetching balances asynchronously:

        >>> from eth_portfolio.protocols import balances
        >>> address = "0x1234567890abcdef1234567890abcdef12345678"
        >>> block = 12345678
        >>> remote_balances = await balances(address, block)
        >>> print(remote_balances)

        Fetching balances synchronously:

        >>> remote_balances = balances(address, block)
        >>> print(remote_balances)

    The function constructs a dictionary `data` with protocol class names
    as keys and their corresponding protocol balances as values. The `protocol_balances`
    variable is a result of mapping the `balances` method over the `protocols` using
    :func:`a_sync.map`. The asynchronous comprehension iterates over `protocol_balances`
    to filter and construct the `data` dictionary. This dictionary is subsequently used
    to initialize the :class:`~eth_portfolio.typing.RemoteTokenBalances` object.
    """
    if not protocols:
        return RemoteTokenBalances(block=block)
    protocol_balances = a_sync.map(
        lambda protocol: protocol.balances(address, block),
        protocols,
    )
    data = {
        type(protocol).__name__: protocol_balances
        async for protocol, protocol_balances in protocol_balances
        if protocol_balances is not None
    }
    return RemoteTokenBalances(data, block=block)



import re
from asyncio import Task, create_task, sleep
from datetime import datetime, timedelta, timezone
from typing import Any, AsyncGenerator, Dict, List, Optional

from brownie import chain


def parse_timedelta(value: str) -> timedelta:
    regex = re.compile(r"(\d+)([dhms]?)")
    result = regex.findall(value)

    days, hours, minutes, seconds = 0, 0, 0, 0

    for val, unit in result:
        val = int(val)
        if unit == "d":
            days = val
        elif unit == "h":
            hours = val
        elif unit == "m":
            minutes = val
        elif unit == "s":
            seconds = val

    return timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds)


async def aiter_timestamps(
    *,
    start: Optional[datetime] = None,
    interval: timedelta = timedelta(days=1),
    run_forever: bool = False,
) -> AsyncGenerator[datetime, None]:
    """
    Generates the timestamps to be queried based on the specified range and interval.
    """
    if not isinstance(run_forever, bool):
        raise TypeError(f"`run_forever` must be boolean. You passed {run_forever}")

    if start is None:
        start = datetime.now(tz=timezone.utc)

    block0_ts = datetime.fromtimestamp(chain[0].timestamp, tz=timezone.utc)
    helper = datetime(
        year=block0_ts.year,
        month=block0_ts.month,
        day=block0_ts.day,
        hour=block0_ts.hour,
        minute=block0_ts.minute,
        tzinfo=timezone.utc,
    )
    while helper + interval < start:
        helper += interval
    start = helper
    if start < block0_ts:
        start += interval

    timestamp = start

    while timestamp <= datetime.now(tz=timezone.utc):
        yield timestamp
        timestamp = timestamp + interval

    while run_forever:
        while timestamp > datetime.now(tz=timezone.utc):
            await _get_waiter(timestamp)
        yield timestamp
        timestamp += interval


_waiters: Dict[datetime, "Task[None]"] = {}


def _get_waiter(timestamp: datetime) -> "Task[None]":
    if timestamp not in _waiters:
        waiter = create_task(sleep_until(timestamp))
        waiter.add_done_callback(_sleep_done_callback)
        _waiters[timestamp] = waiter
    return _waiters[timestamp]


async def sleep_until(until: datetime) -> None:
    now = datetime.now(tz=timezone.utc)
    await sleep((until - now).total_seconds())


def _sleep_done_callback(t: "Task[Any]") -> None:
    low_to_hi = sorted(_waiters)
    for k in low_to_hi:
        if _waiters[k] is t:
            _waiters.pop(k)



import asyncio
from argparse import ArgumentParser
from os import environ

import brownie

from eth_portfolio_scripts import docker, logger
from eth_portfolio_scripts._args import add_infra_port_args
from eth_portfolio_scripts.balances import export_balances


parser = ArgumentParser(description="eth-portfolio")

subparsers = parser.add_subparsers(title="Commands", dest="command", required=True)

export_parser = subparsers.add_parser("export", help="Export a specific dataset for your portfolio")
export_parser.add_argument("target", help="Choose an exporter to run")
add_infra_port_args(export_parser)
export_parser.set_defaults(func=export_balances)

infra_parser = subparsers.add_parser(
    "infra", help="Start the docker containers required to run any eth-portfolio service"
)
infra_parser.add_argument("cmd", help="What do you want to do?")
add_infra_port_args(infra_parser)
infra_parser.add_argument(
    "--start-renderer",
    action="store_true",
    help="If set, starts the renderer container in addition to the default containers. By default, only grafana, victoria-metrics, and vmagent are started.",
)

export_parser.add_argument(
    "--wallet",
    type=str,
    help="The address of a wallet to export. You can pass multiple, ie. `--wallet 0x123 0x234 0x345`",
    required=True,
    nargs="+",
)
export_parser.add_argument(
    "--network",
    type=str,
    help="The brownie network identifier for the rpc you wish to use. default: mainnet",
    default="mainnet",
)
export_parser.add_argument(
    "--label",
    type=str,
    help='The label for this portfolio, if you want one. Defaults to "MyPortfolio".',
    default="My Portfolio",
)
export_parser.add_argument(
    "--interval",
    type=str,
    help="The time interval between datapoints. default: 6h",
    default="6h",
)
export_parser.add_argument(
    "--first-tx-block",
    type=int,
    help=(
        "The block of your portfolio's first transaction, if known. "
        "This value, if provided, allows us to speed up processing of your data by limiting the block range we need to query. "
        "If not provided, the whole blockchain will be scanned."
    ),
    default=0,
)
export_parser.add_argument(
    "--export-start-block",
    type=int,
    help="The first block in the range you wish to export.",
    default=0,
)
export_parser.add_argument(
    "--daemon",
    action="store_true",
    help="TODO: If True, starts a daemon process instead of running in your terminal. Not currently supported.",
)

args = parser.parse_args()

if hasattr(args, "network"):
    environ["BROWNIE_NETWORK_ID"] = args.network

environ["GRAFANA_PORT"] = str(args.grafana_port)
environ["RENDERER_PORT"] = str(args.renderer_port)
environ["VICTORIA_PORT"] = str(args.victoria_port)


# TODO: run forever arg
def main():
    command = args.command
    if command == "infra":
        if args.cmd == "start":
            # Start the backend containers
            if getattr(args, "start_renderer", False):
                docker.up()
            else:
                docker.up("grafana", "victoria-metrics", "vmagent")
        elif args.cmd == "stop":
            docker.down()
        else:
            raise ValueError(f"{args.target} is not a valid command")

    else:
        # The user's command is `export`
        if args.target == "balances":
            asyncio.run(args.func(args))
        else:
            raise ValueError(f"{args.target} is not a valid command")


if __name__ == "__main__":
    brownie.project.run(__file__)



from y import ERC20, Network

from eth_portfolio.protocols._base import ProtocolWithStakingABC, SingleTokenStakingPoolABC


class _CvxLockerV2(SingleTokenStakingPoolABC):
    contract_address = "0x72a19342e8F1838460eBFCCEf09F6585e32db86E"
    balance_method_name = "lockedBalanceOf"
    token = ERC20("0x4e3FBD56CD56c3e72c1403e103b45Db9da5B9D2B")


class Convex(ProtocolWithStakingABC):
    networks = [Network.Mainnet]

    def __init__(self) -> None:
        super().__init__()
        self.pools = [_CvxLockerV2()]



import logging
import sqlite3
from abc import abstractmethod
from datetime import datetime
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    AsyncGenerator,
    AsyncIterator,
    Final,
    Generic,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import dank_mids
from a_sync import ASyncGenericBase, ASyncIterable, ASyncIterator, as_yielded
from async_lru import alru_cache
from brownie import chain
from brownie.exceptions import ContractNotFound
from eth_abi.exceptions import InsufficientDataBytes
from eth_typing import ChecksumAddress
from pandas import DataFrame  # type: ignore
from y import ERC20, Contract, Network
from y.constants import CHAINID, NETWORK_NAME
from y.datatypes import AddressOrContract, Block
from y.exceptions import (
    CantFetchParam,
    ContractNotVerified,
    NodeNotSynced,
    NonStandardERC20,
    PriceError,
    reraise_excs_with_extra_context,
    yPriceMagicError,
)
from y.prices.magic import get_price

from eth_portfolio import _config, _decimal
from eth_portfolio.typing import _T

if TYPE_CHECKING:
    from eth_portfolio.structs import LedgerEntry


logger: Final = logging.getLogger(__name__)

NON_STANDARD_ERC721: Final = {
    Network.Mainnet: ["0xb47e3cd837dDF8e4c57F05d70Ab865de6e193BBB"],  # CryptoPunks
}.get(CHAINID, [])


async def get_buffered_chain_height() -> int:
    """Returns an int equal to the current height of the chain minus `_config.REORG_BUFFER`."""
    return await dank_mids.eth.get_block_number() - _config.REORG_BUFFER


class PandableList(List[_T]):
    @cached_property
    def df(self) -> DataFrame:
        return self._df()

    def _df(self) -> DataFrame:
        """Override this method if you need to manipulate your dataframe before returning it."""
        return DataFrame(self)


class Decimal(_decimal.Decimal):
    """
    I'm in the process of moving from floats to decimals, this will help be as I buidl.
    """

    # TODO i forget why I had this lol, should I delete?

    def __init__(self, v) -> None:
        # assert not isinstance(v, _decimal.Decimal)
        super().__init__()


async def _describe_err(token: AddressOrContract, block: Optional[Block]) -> str:
    """
    Assembles a string used to provide as much useful information as possible in PriceError messages
    """
    try:
        symbol = await ERC20(token, asynchronous=True).symbol
    except NonStandardERC20:
        symbol = None

    if block is None:
        if symbol:
            return f"{symbol} {token} on {NETWORK_NAME}"

        return f"malformed token {token} on {NETWORK_NAME}"

    if symbol:
        return f"{symbol} {token} on {NETWORK_NAME} at {block}"

    return f"malformed token {token} on {NETWORK_NAME} at {block}"


_to_raise: Final = (
    OSError,
    FileNotFoundError,
    NodeNotSynced,
    NotImplementedError,
    sqlite3.OperationalError,
    InsufficientDataBytes,
    UnboundLocalError,
    RuntimeError,
)


async def _get_price(token: AddressOrContract, block: Optional[int] = None) -> _decimal.Decimal:
    token = str(token)
    with reraise_excs_with_extra_context(token, block):
        try:
            if await is_erc721(token):
                return _decimal.Decimal(0)
            maybe_float = await get_price(token, block, silent=True, sync=False)
            dprice = _decimal.Decimal(maybe_float)
            return round(dprice, 18)
        except CantFetchParam as e:
            logger.warning("CantFetchParam %s", e)
        except yPriceMagicError as e:
            # Raise these exceptions
            if isinstance(e.exception, _to_raise) and not isinstance(e.exception, RecursionError):
                raise e.exception
            # The exceptions below are acceptable enough
            elif isinstance(e.exception, NonStandardERC20):
                # Can't get symbol for handling like other excs
                logger.warning(f"NonStandardERC20 while fetching price for {token}")
            elif isinstance(e.exception, PriceError):
                logger.warning(
                    f"PriceError while fetching price for {await _describe_err(token, block)}"
                )
            else:
                logger.warning(f"{e} while fetching price for {await _describe_err(token, block)}")
                logger.warning(e, exc_info=True)
    return _decimal.Decimal(0)


@alru_cache(maxsize=None)
async def is_erc721(token: ChecksumAddress) -> bool:
    # This can probably be improved
    try:
        contract = await Contract.coroutine(token)
    except (ContractNotFound, ContractNotVerified):
        return False
    attrs = "setApprovalForAll", "getApproved", "isApprovedForAll"
    if all(hasattr(contract, attr) for attr in attrs):
        return True
    elif contract.address in NON_STANDARD_ERC721:
        return True
    return False


def _unpack_indicies(indicies: Union[Block, Tuple[Block, Block]]) -> Tuple[Block, Block]:
    """Unpacks indicies and returns a tuple (start_block, end_block)."""
    if isinstance(indicies, tuple):
        start_block, end_block = indicies
    else:
        start_block = 0
        end_block = indicies
    return start_block, end_block


class _AiterMixin(ASyncIterable[_T]):
    def __aiter__(self) -> AsyncIterator[_T]:
        return self[self._start_block : chain.height].__aiter__()

    def __getitem__(self, slice: slice) -> ASyncIterator[_T]:
        if slice.start is not None and not isinstance(slice.start, (int, datetime)):
            raise TypeError(f"start must be int or datetime. you passed {slice.start}")
        if slice.stop and not isinstance(slice.stop, (int, datetime)):
            raise TypeError(f"start must be int or datetime. you passed {slice.start}")
        if slice.step is not None:
            raise ValueError("You cannot use a step here.")
        return ASyncIterator(self._get_and_yield(slice.start or 0, slice.stop, True))  # type: ignore [truthy-function]

    def yield_forever(self, mem_cache: bool = True) -> ASyncIterator[_T]:
        return ASyncIterator(self._get_and_yield(self._start_block or 0, None, mem_cache))

    @abstractmethod
    async def _get_and_yield(
        self, start_block: Block, end_block: Block, mem_cache: bool
    ) -> AsyncGenerator[_T, None]:
        yield

    @property
    @abstractmethod
    def _start_block(self) -> int: ...


_LT = TypeVar("_LT")


class _LedgeredBase(ASyncGenericBase, _AiterMixin["LedgerEntry"], Generic[_LT]):
    """A mixin class for things with ledgers"""

    transactions: _LT
    internal_transfers: _LT
    token_transfers: _LT

    def __init__(self, start_block: Block) -> None:
        self.start_block = start_block
        super().__init__()

    @property
    def _start_block(self) -> Block:
        # in the middle of refactoring this
        return self.start_block

    @cached_property
    def _ledgers(self) -> Tuple[_LT, _LT, _LT]:
        """A tuple with the 3 ledgers (transactions, internal transfers, token transfers)"""
        return self.transactions, self.internal_transfers, self.token_transfers

    def _get_and_yield(
        self, start_block: Block, end_block: Block, mem_cache: bool
    ) -> AsyncGenerator["LedgerEntry", None]:
        return as_yielded(*(ledger._get_and_yield(start_block, end_block, mem_cache) for ledger in self._ledgers))  # type: ignore [return-value,index]



import abc
from asyncio import gather
from typing import List, Optional

import a_sync
from a_sync import igather
from a_sync.a_sync import ASyncFunctionSyncDefault, HiddenMethod
from brownie.network.contract import ContractCall
from y import ERC20, Contract
from y._decorators import stuck_coro_debugger
from y.contracts import contract_creation_block
from y.datatypes import Address, Block

from eth_portfolio._utils import Decimal
from eth_portfolio.typing import Balance, TokenBalances


class ProtocolABC(metaclass=abc.ABCMeta):
    @a_sync.future
    async def balances(self, address: Address, block: Optional[Block] = None) -> TokenBalances:
        return await self._balances(address, block=block)

    @abc.abstractmethod
    async def _balances(self, address: Address, block: Optional[Block] = None) -> TokenBalances: ...


class ProtocolWithStakingABC(ProtocolABC, metaclass=abc.ABCMeta):
    pools: List["StakingPoolABC"]

    @stuck_coro_debugger
    async def _balances(self, address: Address, block: Optional[Block] = None) -> TokenBalances:
        return sum(await igather(pool.balances(address, block) for pool in self.pools))  # type: ignore


class StakingPoolABC(ProtocolABC, metaclass=abc.ABCMeta):
    contract_address: Address
    """
    The address of the staking pool.
    """

    balance_method_name: str
    """
    The name of the method that is used to query a staker's balance in the staking pool.
    """

    @a_sync.future
    async def __call__(self, *args, block: Optional[Block] = None, **_) -> int:
        if _:
            raise ValueError(
                "SingleTokenStakingPoolABC.__call__ does not support keyword arguments"
            )
        return await self.contract_call.coroutine(*args, block_identifier=block)

    @property
    def contract(self) -> Contract:
        return Contract(self.contract_address)

    @property
    def contract_call(self) -> ContractCall:
        return getattr(self.contract, self.balance_method_name)

    @property
    def deploy_block(self) -> Block:
        return contract_creation_block(self.contract_address)

    def should_check(self, block: Optional[Block]) -> bool:
        return block is None or block >= self.deploy_block


class SingleTokenStakingPoolABC(StakingPoolABC, metaclass=abc.ABCMeta):
    """
    Works for any contract that has a view method with the following signature:

    ```
    methodName(address) -> uint256
    ```

    In the above example:
     - ``address`` is the user's address, provided at runtime.
     - ``methodName`` is whatever string you set for the staking pool's ``balance_method_name``.
     - ``uint256`` is the user's ``token`` balance held in ``contract_address`` at the queried block.
    """

    token: ERC20
    """
    The token that is used for staking in this pool.
    """

    @property
    def price(self) -> ASyncFunctionSyncDefault[Block, Decimal]:
        return self.token.price

    @property
    def scale(self) -> HiddenMethod[ERC20, int]:
        return self.token.__scale__

    @stuck_coro_debugger
    async def _balances(self, address: Address, block: Optional[Block] = None) -> TokenBalances:
        balances: TokenBalances = TokenBalances(block=block)
        if self.should_check(block):
            balance = Decimal(await self(address, block=block))  # type: ignore
            if balance:
                scale, price = await gather(self.scale, self.price(block, sync=False))
                balance /= scale  # type: ignore
                balances[self.token.address] = Balance(
                    balance, balance * Decimal(price), token=self.token.address, block=block
                )
        return balances



from typing import Optional

from y import Contract, Network
from y._decorators import stuck_coro_debugger
from y.datatypes import Address, Block

from eth_portfolio._decimal import Decimal
from eth_portfolio._utils import _get_price
from eth_portfolio.protocols.lending._base import LendingProtocolWithLockedCollateral
from eth_portfolio.typing import Balance, TokenBalances

# NOTE: This only works for YFI collateral, must extend before using for other collaterals
yfi = "0x0bc529c00C6401aEF6D220BE8C6Ea1667F6Ad93e"
usdp = "0x1456688345527bE1f37E9e627DA0837D6f08C925"


class UnitXyz(LendingProtocolWithLockedCollateral):
    networks = [Network.Mainnet]

    def __init__(self) -> None:
        self.unitVault = Contract("0xb1cff81b9305166ff1efc49a129ad2afcd7bcf19")
        self.start_block = 11315910

    @stuck_coro_debugger
    async def _balances(self, address: Address, block: Optional[Block] = None) -> TokenBalances:
        balances: TokenBalances = TokenBalances(block=block)
        if block and block < self.start_block:
            return balances
        bal: int = await self.unitVault.collaterals.coroutine(yfi, address, block_identifier=block)
        if bal:
            bal = Decimal(bal) / 10**18
            balances[yfi] = Balance(bal, bal * await _get_price(yfi, block), token=yfi, block=block)
        return balances

    @stuck_coro_debugger
    async def _debt(self, address: Address, block: Optional[Block] = None) -> TokenBalances:
        balances: TokenBalances = TokenBalances(block=block)
        if block and block < self.start_block:
            return balances
        # NOTE: This only works for YFI based debt, must extend before using for other collaterals
        if debt := await self.unitVault.getTotalDebt.coroutine(
            yfi, address, block_identifier=block
        ):
            debt = Decimal(debt) / 10**18
            balances[usdp] = Balance(debt, debt, token=usdp, block=block)
        return balances



from eth_portfolio_scripts.docker.check import check_docker, check_docker_compose, check_system
from eth_portfolio_scripts.docker.docker_compose import build, down, ensure_containers, up, stop

__all__ = [
    # commands
    "build",
    "up",
    "down",
    "stop",
    # decorators
    "ensure_containers",
    # checks
    "check_docker",
    "check_docker_compose",
    "check_system",
]



from os import environ


environ["DANKMIDS_GANACHE_FORK"] = "0"
environ["DANKMIDS_COLLECT_STATS"] = "0"


from dotenv import load_dotenv


load_dotenv()


from eth_portfolio_scripts._logging import logger, setup_logging


setup_logging()


__all__ = ["logger"]



from inspect import getfullargspec
from typing import Any, Callable, List, Tuple, Type

# WIP:


def get_args_type(sample: Callable) -> Tuple[Type, ...]:
    argspec = getfullargspec(sample)
    args = {
        arg_name: argspec.annotations[arg_name] if arg_name in argspec.annotations else Any
        for arg_name in argspec.args
        if arg_name != "self"
    }
    return tuple(*args.values())


def get_kwargs_type(sample: Callable) -> Tuple[Type, ...]:
    argspec = getfullargspec(sample)
    num_kwargs = len(argspec.args) - len(argspec.defaults or [])
    kwarg_names = argspec.args[num_kwargs:]
    kwargs = {
        kwarg_name: argspec.annotations[kwarg_name] if kwarg_name in argspec.annotations else Any
        for kwarg_name in kwarg_names
    }
    _kwarg_types: List[Type[object]] = list(*kwargs.values())
    if num_kwargs == 1:
        return Tuple[_kwarg_types[0]]  # type: ignore [valid-type,return-value]
    elif num_kwargs == 2:
        return Tuple[_kwarg_types[0], _kwarg_types[1]]  # type: ignore [misc,return-value]
    elif num_kwargs == 3:
        return Tuple[_kwarg_types[0], _kwarg_types[1], _kwarg_types[2]]  # type: ignore [misc,return-value]
    else:
        return Any  # type: ignore [misc,return-value]


def get_return_type(sample: Callable) -> Type:
    argspec = getfullargspec(sample)
    return argspec.annotations["return"] if "return" in argspec.annotations else Any


def get_types(sample: Callable) -> Tuple[Type, Type, Type]:
    return get_args_type(sample), get_kwargs_type(sample), get_return_type(sample)  # type: ignore [return-value]



from gzip import compress
from logging import getLogger
from os import environ
from typing import Final, List

import a_sync
from aiohttp import ClientError, ClientSession, ServerDisconnectedError
from msgspec import json

from eth_portfolio_scripts.victoria import types
from eth_portfolio_scripts.victoria.types import Metric


BASE_URL: Final = environ.get("VM_URL", "http://127.0.0.1:8430")

# this will be populated later
session: ClientSession = None  # type: ignore [assignment]

logger: Final = getLogger("eth_portfolio.victoria")

encode: Final = json.encode
decode: Final = json.decode


class VictoriaMetricsError(ValueError): ...


def get_session() -> ClientSession:
    sesh = session
    if sesh is None:
        sesh = ClientSession(BASE_URL, raise_for_status=True)
        __set_session(sesh)
    return sesh


async def get(url: str) -> bytes:
    session = get_session()
    while True:
        try:
            async with session.get(url=url, headers={"Connection": "close"}) as response:
                return await response.read()
        except ServerDisconnectedError:
            continue


@a_sync.Semaphore(2)
async def post_data(metrics_to_export: List["Metric"]) -> None:
    """Post all metrics at once."""
    data = compress(b"\n".join(encode(metric) for metric in metrics_to_export))
    attempts = 0
    session = get_session()
    while True:
        try:
            async with session.post(
                "/api/v1/import",
                headers={"Connection": "close", "Content-Encoding": "gzip"},
                data=data,
            ):
                logger.debug(f"posted {len(data)} datas")
                return
        except ClientError as e:
            attempts += 1
            logger.debug("You had a ClientError: %s", e)
            if attempts >= 10:
                raise e


def __set_session(sesh: ClientSession) -> None:
    global session
    session = sesh



import os
from typing import Final

from brownie import chain
from y import Network, convert, weth

ERC20_TRANSFER_EVENT_HASH: Final = (
    "0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef"
)
ERC677_TRANSFER_EVENT_HASH: Final = (
    "0xe19260aff97b920c7df27010903aeb9c8d2be5d310a2c67824cf3f15396e4c16"
)

NUM_WALLETS: Final = int(os.environ.get("NUM_WALLETS", "10"))
ADDRESSES: Final = [
    convert.to_address(address)
    for address in {os.environ.get(f"PORTFOLIO_ADDRESS_{i}", None) for i in range(NUM_WALLETS)}
    if address is not None
]

TRANSFER_SIGS: Final = [ERC20_TRANSFER_EVENT_HASH, ERC677_TRANSFER_EVENT_HASH]

ETH_LIKE: Final = {
    Network.Mainnet: {
        "ETH",
        weth.address,
        "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE",
        "0x5e74C9036fb86BD7eCdcb084a0673EFc32eA31cb",  # seth
        "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE",  # eth
        "0xae7ab96520DE3A18E5e111B5EaAb095312D7fE84",  # steth
        "0x9559Aaa82d9649C7A7b220E7c461d2E74c9a3593",  # reth
        "0xE95A203B1a91a908F9B9CE46459d101078c2c3cb",  # ankreth
    },
}.get(chain.id, set())

BTC_LIKE: Final = {
    Network.Mainnet: {
        "0xEB4C2781e4ebA804CE9a9803C67d0893436bB27D",  # renbtc
        "0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599",  # wbtc
        "0xfE18be6b3Bd88A2D2A7f928d00292E7a9963CfC6",  # sbtc
        "0x8064d9Ae6cDf087b1bcd5BDf3531bD5d8C537a68",  # obtc
        "0x9BE89D2a4cd102D8Fecc6BF9dA793be995C22541",  # bbtc
        "0x0316EB71485b0Ab14103307bf65a021042c6d380",  # hbtc
        "0x5228a22e72ccC52d415EcFd199F99D0665E7733b",  # pbtc
        "0x8dAEBADE922dF735c38C80C7eBD708Af50815fAa",  # tbtc
    },
}.get(chain.id, set())

INTL_STABLECOINS: Final = {
    Network.Mainnet: {
        "0xD71eCFF9342A5Ced620049e616c5035F1dB98620",  # sEUR
        "0xC581b735A1688071A1746c968e0798D642EDE491",  # EURT
        "0xdB25f211AB05b1c97D595516F45794528a807ad8",  # EURS
        "0x96E61422b6A9bA0e068B6c5ADd4fFaBC6a4aae27",  # ibEUR
        "0x269895a3dF4D73b077Fc823dD6dA1B95f72Aaf9B",  # sKRW
        "0x3F1B0278A9ee595635B61817630cC19DE792f506",  # sAUD
        "0x97fe22E7341a0Cd8Db6F6C021A24Dc8f4DAD855F",  # sGBP
        "0xF6b1C627e95BFc3c1b4c9B825a032Ff0fBf3e07d",  # sJPY
        "0x0F83287FF768D1c1e17a42F44d644D7F22e8ee1d",  # sCHF
        "0x269895a3dF4D73b077Fc823dD6dA1B95f72Aaf9B",  # sKRW
    },
}.get(chain.id, set())



from typing import Final

import dank_mids
import eth_retry
import msgspec
from a_sync import SmartProcessingQueue
from async_lru import alru_cache
from eth_typing import HexStr
from y._decorators import stuck_coro_debugger


Raw: Final = msgspec.Raw
TxReceiptQueue = SmartProcessingQueue[HexStr, [], msgspec.Raw]


@eth_retry.auto_retry(min_sleep_time=1, max_sleep_time=3, max_retries=20, suppress_logs=1)
@alru_cache(maxsize=None, ttl=60 * 60)
@stuck_coro_debugger
async def _get_transaction_receipt(txhash: HexStr) -> msgspec.Raw:
    """
    Fetches the transaction receipt for a given transaction hash.

    This function retrieves the transaction receipt from the Ethereum network
    using the provided transaction hash. It utilizes caching to store results
    for a specified time-to-live (TTL) to improve performance and reduce
    network calls. The function is also decorated to automatically retry
    in case of failures and to debug if the coroutine gets stuck.

    Args:
        txhash: The transaction hash for which to retrieve the receipt.

    Returns:
        msgspec.Raw: The raw transaction receipt data.

    Examples:
        >>> txhash = "0x1234567890abcdef..."
        >>> receipt = await _get_transaction_receipt(txhash)
        >>> print(receipt)

    See Also:
        - :func:`eth_retry.auto_retry`: For automatic retry logic.
        - :func:`async_lru.alru_cache`: For caching the results.
        - :func:`y._decorators.stuck_coro_debugger`: For debugging stuck coroutines.
    """
    return await __get_transaction_receipt(txhash, decode_to=Raw, decode_hook=None)


get_transaction_receipt: Final[TxReceiptQueue] = SmartProcessingQueue(
    _get_transaction_receipt, 5000
)
"""
A queue for processing transaction receipt requests.

This queue manages the processing of transaction receipt requests, allowing
up to 5000 concurrent requests. It uses the `_get_transaction_receipt` function
to fetch the receipts.

Examples:
    >>> txhash = "0x1234567890abcdef..."
    >>> receipt = await get_transaction_receipt(txhash)
    >>> print(receipt)

See Also:
    - :class:`a_sync.SmartProcessingQueue`: For managing asynchronous processing queues.
"""


__get_transaction_receipt: Final = dank_mids.eth.get_transaction_receipt
